from typing import Any, Optional

class PSS_SigScheme:
    def __init__(self, key, mgfunc, saltLen) -> None: ...
    def can_sign(self): ...
    def sign(self, mhash): ...
    def verify(self, mhash, S): ...

def new(key, mgfunc: Optional[Any] = ..., saltLen: Optional[Any] = ...): ...
